from dataclasses import dataclass
from typing import (
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    overload,
)

from sqlalchemy import delete, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from pydantic import BaseModel

from ..decorator.db_async_session import db_async_session

from ..models.decorated_base import DecoratedBase
from ..exceptions import *


# Generic Type for Pydantic and SQLAlchemy
EntityModel = TypeVar("EntityModel", bound=DecoratedBase)
EntitySchema = TypeVar("EntitySchema", bound=BaseModel)


@dataclass
class Repository(Generic[EntityModel, EntitySchema]):
    model: Type[EntityModel]

    @db_async_session
    async def create(
        self,
        data: EntitySchema,
        session: Optional[AsyncSession] = None,
    ) -> EntityModel:
        """Accepts a Pydantic model, creates a new record in the database, catches
        any integrity errors, and returns the record.

        Args:
            data (EntitySchema): Pydantic model
            session (Optional[AsyncSession]): SQLAlchemy async session

        Raises:
            IntegrityConflictException: if creation conflicts with existing data
            EntityException: if an unknown error occurs

        Returns:
            EntityModel: created SQLAlchemy model
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        try:
            db_model = self.model(**data.model_dump())
            session.add(db_model)
            await session.commit()
            await session.refresh(db_model)
            return db_model
        except IntegrityError:
            raise IntegrityConflictException(
                f"{self.model.__tablename__} conflicts with existing data.",
            )
        except Exception as e:
            raise EntityException(f"Unknown error occurred: {e}") from e

    @db_async_session
    async def create_many(
        self,
        data: List[EntitySchema],
        session: Optional[AsyncSession] = None,
    ) -> List[EntityModel]:
        """_summary_

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            data (list[EntitySchema]): list of Pydantic models

        Raises:
            IntegrityConflictException: if creation conflicts with existing data
            EntityException: if an unknown error occurs

        Returns:
            list[EntityModel] | bool: list of created SQLAlchemy models or boolean
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        db_models = [self.model(**d.model_dump()) for d in data]
        if not data:
            return db_models
        try:
            session.add_all(db_models)
            await session.commit()
        except IntegrityError:
            raise IntegrityConflictException(
                f"{self.model.__tablename__} conflict with existing data.",
            )
        except Exception as e:
            raise EntityException(f"Unknown error occurred: {e}") from e
        for m in db_models:
            await session.refresh(m)

        return db_models

    @db_async_session
    async def get_one_by_id(
        self,
        id_: str,
        column: str = "id",
        with_for_update: bool = False,
        session: Optional[AsyncSession] = None,
    ) -> Optional[EntityModel]:
        """Fetches one record from the database based on a column value and returns
        it, or returns None if it does not exist. Raises an exception if the column
        doesn't exist.

        Args:
            session (Optional[AsyncSession], optional): SQLAlchemy async session
            id_ (str): value to search for in `column`.
            column (str, optional): the column name in which to search.
                Defaults to "uuid".
            with_for_update (bool, optional): Should the returned row be locked
                during the lifetime of the current open transactions.
                Defaults to False.

        Raises:
            EntityException: if the column does not exist on the model

        Returns:
            EntityModel: SQLAlchemy model or None
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        try:
            q = select(self.model).where(getattr(self.model, column) == id_)
        except AttributeError:
            raise EntityException(
                f"Column {column} not found on {self.model.__tablename__}.",
            )

        if with_for_update:
            q = q.with_for_update()

        results = await session.execute(q)
        return results.unique().scalar_one_or_none()

    @db_async_session
    async def get_many_by_ids(
        self,
        ids: Optional[List[str]],
        column: str = "id",
        with_for_update: bool = False,
        session: Optional[AsyncSession] = None,
    ) -> List[EntityModel]:
        """Fetches multiple records from the database based on a column value and
        returns them. Raises an exception if the column doesn't exist.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            ids (list[str], optional): list of values to search for in
                `column`. Defaults to None.
            column (str, optional): the column name in which to search
                Defaults to "id".
            with_for_update (bool, optional): Should the returned rows be locked
                during the lifetime of the current open transactions.
                Defaults to False.

        Raises:
            EntityException: if the column does not exist on the model
            NotExistSessionException: if not session append to the calling this method

        Returns:
            list[EntityModel]: list of SQLAlchemy models
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        q = select(self.model)
        if ids:
            try:
                q = q.where(getattr(self.model, column).in_(ids))
            except AttributeError:
                raise EntityException(
                    f"Column {column} not found on {self.model.__tablename__}.",
                )

        if with_for_update:
            q = q.with_for_update()

        rows = await session.execute(q)
        return list(rows.unique().scalars().all())

    @db_async_session
    async def update_entity(
        self,
        entity: EntityModel,
        session: Optional[AsyncSession] = None,
    ) -> None:
        """
        Update a model which is bound to a record in the database.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            model (EntityModel): updated SQLAlchemy mode
        Raises:
            NotFoundException: if the record isn't found
            IntegrityConflictException: if the update conflicts with existing data

        Returns:
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        await session.merge(entity)
        await session.commit()

    @db_async_session
    async def update_by_id(
        self,
        data: EntitySchema,
        id_: str,
        column: str = "id",
        session: Optional[AsyncSession] = None,
    ) -> EntityModel:
        """Updates a record in the database based on a column value and returns the
        updated record. Raises an exception if the record isn't found or if the
        column doesn't exist.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            data (EntitySchema): Pydantic schema for the updated data.
            id_ (str | UUID): value to search for in `column`
            column (str, optional): the column name in which to search
                Defaults to "uuid".
        Raises:
            NotFoundException: if the record isn't found
            IntegrityConflictException: if the update conflicts with existing data

        Returns:
            EntityModel: updated SQLAlchemy model
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        db_model = await self.get_one_by_id(id_, column, True)
        if not db_model:
            raise NotFoundException(
                f"{self.model.__tablename__} {column}={id_} not found.",
            )

        values = data.model_dump(exclude_unset=True)
        for k, v in values.items():
            setattr(db_model, k, v)
        session.add(db_model)

        try:
            await session.commit()
            return db_model
        except IntegrityError:
            raise IntegrityConflictException(
                f"{self.model.__tablename__} {column}={id_} conflict with existing data.",
            )

    @db_async_session
    async def update_many_by_ids(
        self,
        updates: Dict[str, EntitySchema],
        column: str = "id",
        session: Optional[AsyncSession] = None,
    ) -> List[EntityModel]:
        """Updates multiple records in the database based on a column value and
        returns the updated records. Raises an exception if the column doesn't
        exist.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            updates (dict[str  |  UUID, EntitySchema]): dictionary of id_ to
                Pydantic update schema
            column (str, optional): the column name in which to search.
                Defaults to "uuid".
            return_models (bool, optional): Should the created models be returned
                or a boolean indicating they have been created. Defaults to False.
                Defaults to False.

        Raises:
            IntegrityConflictException: if the update conflicts with existing data

        Returns:
            list[EntityModel] | bool: list of updated SQLAlchemy models or boolean
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        updates = {str(id): update for id, update in updates.items() if update}
        ids = list(updates.keys())
        db_models = await self.get_many_by_ids(
            ids=ids, column=column, with_for_update=True
        )

        for db_model in db_models:
            values = updates[str(getattr(db_model, column))].model_dump(
                exclude_unset=True
            )
            for k, v in values.items():
                setattr(db_model, k, v)
            session.add(db_model)

        try:
            await session.commit()
        except IntegrityError:
            raise IntegrityConflictException(
                f"{self.model.__tablename__} conflict with existing data.",
            )
        return db_models

    @db_async_session
    async def remove_by_id(
        self,
        id_: str,
        column: str = "id",
        session: Optional[AsyncSession] = None,
    ) -> int:
        """Removes a record from the database based on a column value. Raises an
        exception if the column doesn't exist.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            id (str | UUID): value to search for in `column` and delete
            column (str, optional): the column name in which to search.
                Defaults to "uuid".

        Raises:
            EntityException: if the column does not exist on the model

        Returns:
            int: number of rows removed, 1 if successful, 0 if not. Can be greater
                than 1 if id_ is not unique in the column.
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        try:
            query = delete(self.model).where(getattr(self.model, column) == id_)
        except AttributeError:
            raise EntityException(
                f"Column {column} not found on {self.model.__tablename__}.",
            )

        rows = await session.execute(query)
        await session.commit()
        return rows.rowcount

    @db_async_session
    async def remove_many_by_ids(
        self,
        ids: List[str],
        column: str = "id",
        session: Optional[AsyncSession] = None,
    ) -> int:
        """Removes multiple records from the database based on a column value.
        Raises an exception if the column doesn't exist.

        Args:
            session (Optional[AsyncSession]): SQLAlchemy async session
            ids (list[str  |  UUID]): list of values to search for in `column` and
            column (str, optional): the column name in which to search.
                Defaults to "uuid".

        Raises:
            EntityException: if ids is empty to stop deleting an entire table
            EntityException: if column does not exist on the model

        Returns:
            int: _description_
        """
        if not session:
            raise NotExistedSessionException("session is not existed")
        if not ids:
            raise EntityException("No ids provided.")

        try:
            query = delete(self.model).where(getattr(self.model, column).in_(ids))
        except AttributeError:
            raise EntityException(
                f"Column {column} not found on {self.model.__tablename__}.",
            )

        rows = await session.execute(query)
        await session.commit()
        return rows.rowcount



__all__ = [
    "Asset",
    "DataType",
    "Exchange",
    "IndicatorType",
    "Market",
    "OrderSide",
    "OrderStatus",
    "OrderType",
    "PositionSide",
    "PositionStatus",
    "MarketStat",
]

from .asset import Asset
from .data_types import DataType
from .exchanges import Exchange
from .indicators import IndicatorType
from .markets import Market
from .order_side import OrderSide
from .order_status import OrderStatus
from .order_type import OrderType
from .position_side import PositionSide
from .position_status import PositionStatus
from .market_stat import MarketStat



from datetime import datetime
from sqlalchemy.orm import Mapped, mapped_column

from ..helpers.get_current_time import GetCurrentTime
from ..models.decorated_base import DecoratedBase


class DatetimeDecoratedBase(DecoratedBase):
    __abstract__ = True
    updated_at: Mapped[datetime] = mapped_column(
        index=True,
        doc="Last Update Time",
        default=GetCurrentTime.get,
        onupdate=GetCurrentTime.get,
    )
    created_at: Mapped[datetime] = mapped_column(
        index=True, doc="Creation Time", default=GetCurrentTime.get
    )



from typing import Any, Dict
import os
import redis.asyncio as aioredis
from redis.asyncio import Redis


class RedisClient:
    """RedisClient.
    This class is going to create redis connection client.
    """

    redis: Redis

    def __init__(self, redis: Redis):
        self.redis = redis

    @classmethod
    async def create(
        cls,
        host: str | None = None,
        port: int | None = None,
        username: str | None = None,
        password: str | None = None,
    ):
        """create.
        this method create a redis client connection
        You can put these input arguments in the .env file and use dotenv
        in order to load .env file for connecting to the Redis.

        Args:
            host (str): host
            port (int): port
            username (str): username
            password (str): password
        """
        if not host:
            host = os.getenv("REDIS_HOST", "localhost")
        if not port:
            port = int(os.getenv("REDIS_PORT", 6379))
        if not username:
            username = os.getenv("REDIS_USERNAME", "")
        if not password:
            password = os.getenv("REDIS_PASSWORD", "")

        if password:
            if username:
                url = f"redis://{username}:{password}@{host}:{port}"
            else:
                url = f"redis://default:{password}@{host}:{port}"
        else:
            url = f"redis://{host}:{port}"

        redis = await aioredis.from_url(
            url,
            decode_responses=True,
            socket_timeout=5,  # timeout for read/write ops
            socket_connect_timeout=5,  # timeout for initial connection
            retry_on_timeout=True,
            health_check_interval=30,
        )
        return cls(redis)

    async def close(self):
        """close.
        close redis connection
        """
        await self.redis.aclose()



import logging
import coloredlogs
from ..decorator.singleton import singleton


LOG_FORMAT = "%(asctime)s [%(levelname)s] %(name)s [%(funcName)s]: %(message)s"

LEVEL_STYLES = {
    "debug": {"color": "cyan"},
    "info": {"color": "green"},
    "warning": {"color": "yellow"},
    "error": {"color": "red"},
    "critical": {"color": "white", "bold": True, "background": "red"},
}

FIELD_STYLES = {
    "asctime": {"color": "magenta"},
    "name": {"color": "blue"},
    "funcName": {"color": "cyan", "bold": True},
}


@singleton
class LoggerFactory:
    def __init__(self, level: str = "INFO"):
        self.level = level.upper()

        # Configure root logger only once
        root_logger = logging.getLogger()
        if not root_logger.hasHandlers():
            coloredlogs.install(
                level=self.level,
                logger=root_logger,
                fmt=LOG_FORMAT,
                datefmt="%H:%M:%S",
                field_styles=FIELD_STYLES,
                level_styles=LEVEL_STYLES,
            )

    def get(self, name: str = __name__) -> logging.Logger:
        return logging.getLogger(name)



from pydantic import BaseModel, Field

from ..enums.exchanges import Exchange
from ..enums.markets import Market
from ..enums.data_types import DataType

from typing import List, Dict, Any, Union, Annotated, Literal


# --- Market Subscribe ---
class MarketSubscriptionBase(BaseModel):
    exchange: Exchange
    market: Market
    data_type: DataType


class NonCandleSubscription(MarketSubscriptionBase):
    data_type: Literal[DataType.TRADES, DataType.ORDERBOOK]


class CandleSubscription(MarketSubscriptionBase):
    data_type: Literal[DataType.CANDLE]
    timeframe: Literal["1m", "5m"]


MarketSubscriptionRequestSchema = Annotated[
    Union[CandleSubscription, NonCandleSubscription],
    Field(discriminator="data_type"),
]



class RedisChannelSubException(Exception):
    pass


class EntityException(Exception):
    pass


class IntegrityConflictException(Exception):
    pass


class NotFoundException(Exception):
    pass


class NotExistedSessionException(Exception):
    pass



__all__ = [
    "IndicatorSubscriptionRequest",
    "MarketSubscriptionRequestSchema",
    "SubscriptionResponseSchema",
    "CandleResponseSchema",
    "PublishDataSchema",
]

from .indicators_schema import IndicatorSubscriptionRequest
from .market_data_schemas import MarketSubscriptionRequestSchema
from .responses_schema import *



import numpy as np
from multiprocessing.shared_memory import SharedMemory
from typing import Any, Dict, List

from ..enums import Market
from ..enums import MarketStat


class MonitoringSHMRepository:
    stat_name: str = "market_stat"
    stats: np.ndarray
    stats_length: int
    stats_sm: SharedMemory
    close_prices_name: str = "close_prices_data"
    close_prices: np.ndarray
    close_prices_length: int = 200
    close_prices_sm: SharedMemory
    reader: bool

    row_index: Dict[Market, int]

    def __init__(
        self, create: bool = False, markets: List[Market] = [Market.BTCUSD_PERP]
    ) -> None:
        # init row index
        self.row_index = dict()
        for i in range(len(markets)):
            self.row_index[markets[i]] = i
        self.stats_length = MarketStat.__len__()

        if create:
            self.reader = False
            try:
                self.create_stat_shm()
            except FileExistsError:
                self.connect_stat_shm()
                self.close_stat()
                self.create_stat_shm()
            try:
                self.create_close_prices_shm()
            except FileExistsError:
                self.connect_close_prices_shm()
                self.close_close_prices()
                self.create_close_prices_shm()
        else:
            self.reader = True
            self.connect_stat_shm()
            self.connect_close_prices_shm()

        # access to arrays
        self.stats = np.ndarray(
            shape=(len(markets), self.stats_length),
            dtype=np.double,
            buffer=self.stats_sm.buf,
        )
        self.close_prices = np.ndarray(
            shape=(len(markets), self.close_prices_length),
            dtype=np.double,
            buffer=self.close_price_sm.buf,
        )
        # initial value
        if create:
            self.stats.fill(0)
            self.close_prices.fill(0)

    def create_stat_shm(self) -> None:
        stat_size = len(self.row_index) * self.stats_length * 8
        self.stats_sm = SharedMemory(name=self.stat_name, create=True, size=stat_size)

    def create_close_prices_shm(self) -> None:
        close_prices_size = len(self.row_index) * self.close_prices_length * 8
        self.close_price_sm = SharedMemory(
            name=self.close_prices_name, create=True, size=close_prices_size
        )

    def connect_stat_shm(self) -> None:
        self.stats_sm = SharedMemory(name=self.stat_name)

    def connect_close_prices_shm(self) -> None:
        self.close_price_sm = SharedMemory(name=self.close_prices_name)

    def close(self) -> None:
        self.close_close_prices()
        self.close_stat()

    def close_close_prices(self) -> None:
        self.close_price_sm.close()
        self.close_price_sm.unlink()

    def close_stat(self) -> None:
        self.stats_sm.close()
        self.stats_sm.unlink()

    def get_close_prices(self, market: Market) -> np.ndarray:
        return self.close_prices[self.row_index[market]]

    def set_close_prices(self, market: Market, close_prices: np.ndarray) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.close_prices[self.row_index[market]] = close_prices

    def get_stat(self, market: Market, stat: MarketStat) -> Any:
        return self.stats[self.row_index[market]][stat.value]

    def set_stat(self, market: Market, stat: MarketStat, value: Any) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.stats[self.row_index[market]][stat.value] = value

    def get_current_candle(self, market: Market) -> Any:
        return self.stats[self.row_index[market]][MarketStat.CANDLE_TIME.value]

    def set_current_candle(self, market: Market, value: Any) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.stats[self.row_index[market]][MarketStat.CANDLE_TIME.value] = value

    def get_last_trade(self, market: Market) -> Any:
        return self.stats[self.row_index[market]][MarketStat.PRICE.value]

    def set_last_trade(self, market: Market, value: Any) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.stats[self.row_index[market]][MarketStat.PRICE.value] = value

    def is_updated(self, market: Market) -> bool:
        return bool(self.stats[self.row_index[market]][MarketStat.IS_UPDATED.value])

    def set_is_updated(self, market: Market) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.stats[self.row_index[market]][MarketStat.IS_UPDATED.value] = 1

    def clear_is_updated(self, market: Market) -> None:
        if self.reader:
            raise Exception("Reader couldn't set the value!!!")
        self.stats[self.row_index[market]][MarketStat.IS_UPDATED.value] = 0



from typing import List, Dict, Any, Literal
from pydantic import BaseModel

from ..enums.data_types import DataType


# --- Response Schema ---
class SubscriptionResponseSchema(BaseModel):
    channel: str


class CandleResponseSchema(BaseModel):
    type: str
    response: List[Dict[str, Any]]


class PublishDataSchema(BaseModel):
    data: dict
    type: DataType
    timeframe: Literal["1m", "5m"] = "1m"



import asyncio
import traceback
import threading
import multiprocessing

from multiprocessing.synchronize import Event
from abc import ABC, abstractmethod
from ..helpers.get_logger import LoggerFactory


LOGGER = LoggerFactory().get(__name__)


class BaseEngine(ABC):
    """
    Abstract base class for a threaded asynchronous engine using its own event loop.

    This class provides a structured lifecycle for a engine that runs an asyncio event loop
    in a separate thread. Subclasses must implement the `preprocess`, `process`, and
    `postprocess` asynchronous methods to define their behavior.

    Attributes:
        name (str): The name of the engine.
        thread_name (str): Name of the thread running the event loop.
        new_loop (asyncio.AbstractEventLoop): The asyncio event loop for this engine.
        thread (Optional[threading.Thread]): The thread in which the event loop runs.
    """

    name: str
    loop_name: str

    def __init__(self, run_in_process: bool = False):
        """
        Initializes the engine instance by creating a new asyncio event loop
        and setting the thread placeholder to None.

        Args:
            run_in_process (bool): If True, run the engine in a multiprocessing.Process
                                instead of a threading.Thread. Defaults to False.
        """
        self.run_in_process = run_in_process
        self.new_loop = asyncio.new_event_loop()
        # can be Thread or Process depending on multi_process flag
        # stop_process_event only used in process mode
        self.therad = None
        self.process = None
        self.stop_process_event = None
        self.pipeline_task = None

    def start(self):
        """
        Starts the engine by running the `preprocess()` coroutine, launching a new thread
        to host the event loop, and scheduling the `process()` coroutine within that loop.
        """
        LOGGER.info(f"starting {self.name}....")
        if self.run_in_process:
            self.stop_process_event = multiprocessing.Event()
            self.process = multiprocessing.Process(
                target=self.start_process_loop, args=(self.stop_process_event,)
            )
            self.loop_name = self.process.name
            self.process.start()
        else:
            self.thread = threading.Thread(target=self.start_thread_loop)
            self.loop_name = self.thread.name
            self.thread.start()

    def start_thread_loop(self):
        """
        Runs the event loop in the current thread.

        This method is intended to be run in a new thread, and will set the thread-local
        event loop and run it until stopped.
        Logs when the loop is stopped.
        """
        asyncio.set_event_loop(self.new_loop)
        try:
            self.pipeline_task = self.new_loop.create_task(self.pipeline())
            self.new_loop.run_forever()
        except Exception as ex:
            msg_error = traceback.format_exc()
            LOGGER.error(f"engine crash {msg_error}")
        LOGGER.info(
            f"Event loop of engine {self.name} in thread {self.loop_name} stopped"
        )

    def start_process_loop(self, stop_process_event: Event):
        """
        Runs the event loop in the current thread.

        This method is intended to be run in a new thread, and will set the thread-local
        event loop and run it until stopped.
        Logs when the loop is stopped.
        """
        self.new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.new_loop)

        async def runner():
            # Start main process task
            self.pipeline_task = asyncio.create_task(self.pipeline())

            # Wait for stop_process_event signal
            while not stop_process_event.is_set():
                await asyncio.sleep(0.2)

            self.pipeline_task.cancel()  # cancel main loop if still running
            self.new_loop.stop()

        try:
            self.new_loop.create_task(runner())
            self.new_loop.run_forever()
        except Exception:
            msg_error = traceback.format_exc()
            LOGGER.error(f"engine crash {msg_error}")
        finally:
            LOGGER.info(
                f"Event loop of engine {self.name} in process {self.loop_name} stopped"
            )

    async def pipeline(self):
        await self.prepare()
        try:
            await self.execute()
        finally:
            await self.postpare()

    @abstractmethod
    async def prepare(self):
        """
        Coroutine for performing setup tasks before the main engine logic begins.

        Must be implemented by subclasses.
        """
        pass

    @abstractmethod
    async def execute(self):
        """
        Coroutine that contains the main logic of the engine.

        This is the task that runs in the engine's event loop and should
        keep the engine alive as long as it's active.

        Must be implemented by subclasses.
        """
        pass

    @abstractmethod
    async def postpare(self):
        """
        Coroutine for performing cleanup tasks after the engine is stopped.

        Must be implemented by subclasses.
        """
        pass

    async def stop(self):
        """
        Stops the engine gracefully.

        Calls the `postprocess()` coroutine, schedules the event loop to stop,
        and joins the engine thread to ensure a clean shutdown.
        """
        await self.stop_process() if self.run_in_process else await self.stop_thread()

    async def stop_process(self):
        """
        Stops the process gracefully.

        Calls the `postprocess()` coroutine, schedules the event loop to stop,
        and joins the engine thread to ensure a clean shutdown.
        """
        if not self.process or not self.stop_process_event:
            return
        # signal process to shut down gracefully by running postprocess()
        # we cant use self.worker.terminate() cause we are running in child process
        # we have to use events to send shutdown signal
        self.stop_process_event.set()
        self.process.join()
        self.process = None
        self.stop_process_event = None

    async def stop_thread(self):
        if not self.thread or not self.pipeline_task:
            return
        self.pipeline_task.cancel()
        self.new_loop.call_soon_threadsafe(self.new_loop.stop)
        self.thread.join()
        self.thread = None
